/* autogenerated by Processing revision 1283 on 2023-06-10 */
import processing.core.*;
import processing.data.*;
import processing.event.*;
import processing.opengl.*;

import codeanticode.syphon.*;
import oscP5.*;
import netP5.*;

import java.util.HashMap;
import java.util.ArrayList;
import java.io.File;
import java.io.BufferedReader;
import java.io.PrintWriter;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;

public class CELLULAR_AUTOMATA extends PApplet {

/*

CELLULAR AUTOMATA GENERATOR
a tool for generating images and videos
based on Cellular Automata concept.
change rules and colour palettes on the go,
add your images, use brushes and share
the results with other sketches or applications
via Syphon. sync to music for live performances!

Ksawery Kirklewski
www.ksawerykomputery.pl/tools/ca-generator
5-01-2020

*/

// Syphon

SyphonServer server;

// Sets
int WindowW, WindowH;
String modeSet;
int scale = 0;
int maxZoom = 64;

boolean sendSyphon = true;
boolean automator = true;

 public void setup() {
  /* size commented out by preprocessor */;
  /* noSmooth commented out by preprocessor */;
  frameRate(30);

  hint(DISABLE_TEXTURE_MIPMAPS);
  ((PGraphicsOpenGL)g).textureSampling(2);
  
  WindowW = 512;
  WindowH = 512;
  
  modeSet = "RULES";
  
  /* default */
  ruleSet = "Bob";
  brushSet = "Rect";  
  initScreen(scale);
  loadPaletteThumbs();  
  setPalette("Greyscale");
  loadImages();
  loadBrushes();
  initConsole();
  if(sendSyphon) server = new SyphonServer(this, "CA");
  initScreen(2);  
  if(automator) initAutomator();
}

 public void draw() {
  background(bgColor);
  CA();
  drawing();
  updateConsole();
  rec();
  if(sendSyphon) server.sendImage(screen);
}
 public void CA() {

  for (int y=0; y<H; y++) {
    for (int x=0; x<W; x++) {
      buffer[x][y] = checkRuleState(x, y);
    }
  }

  output.beginDraw();
  for (int y=0; y<H; y++) {
    for (int x=0; x<W; x++) {
      output.set(x, y, color( buffer[x][y] ) );
      screen.set(x, y, palette[ constrain(buffer[x][y], 0, 255) ] );
    }
  }
  output.endDraw();
  
  pushMatrix();
    translate(margin, margin);
    //noFill();
    //stroke(40);
    //rect(0,0,WindowW,WindowH);
    scale(scale);
    image(screen, 0, 0);
  popMatrix();
  
}
PGraphics screenshot;

 public void keyPressed() {
  
  automatorListen();
  
  if( key != '\\' && key != '[' && key != ']' ) { screenSave(outputTemp); }
  
  switch(key) {
    
    case '1': modeSet = "RULES"; break;
    case '2': modeSet = "PALETTES"; break;
    case '3': modeSet = "IMAGES"; break;
    case '4': modeSet = "BRUSHES"; break;
    case '5': modeSet = "GLITCHES"; break;
    
    case '7': automatorOn(); break;
    case '8': clearAutomator(); break;
    case '9': eraserAutomator(); break;
    
    case '0': recOn(); break;
    case '.': loadPerlinMap(); break;
    case ' ': pauseScreen(); break;
    
    case '-': initScreen(0.5f); break;
    case '=': initScreen(2); break;
    
    case '[': screenSave(outputClipboard); break;
    case ']': screenPaste(outputClipboard); break;
    case '\\': screenPaste(outputTemp); break;
    
    case BACKSPACE: clearScreen(); break;
    case ENTER: screenSaveToFile(); break;

  }
  
  
  switch(modeSet) {
    
    case "RULES":
      for(int i=0; i<RULES.length; i++) {
        if(key==RULES[i].shortcut) { setRule(RULES[i].name); break; }
      }
      break;
  
    case "PALETTES":
      for(int i=0; i<PALETTES.length; i++) {
        if(key==PALETTES[i].shortcut) { setPalette(PALETTES[i].name); PALETTES[i].createThumb(); break; }
      }
      break;
    
    case "BRUSHES":
      for(int i=0; i<BRUSHES.length; i++) {
        if(key==BRUSHES[i].shortcut) { setBrush(BRUSHES[i].name); break; }
      }
      break;
    
    case "IMAGES":
      for(int i=0; i<IMAGES.length; i++) {
        if(key==IMAGES[i].shortcut) { addImage(IMAGES[i].image); break; }
      }
      break;
    
    case "GENERATE":
      for(int i=0; i<GENERATE.length; i++) {
        if(key==GENERATE[i].shortcut) { addElement(GENERATE[i].name); break; }
      }
      break;
      
  }
}


 public void mousePressed() {
  screenSave(outputTemp);
  automatorListen();
}
PImage screen;
PGraphics output;

PGraphics outputTemp;
PGraphics outputClipboard;

int[][] buffer;
int W, H;
int margin;

 public void initScreen(float newscale) {
    
  if(scale==0) {
    
    scale = 1;
    W = WindowW;
    H = WindowH;
    
    output = createGraphics(W, H);
    output.beginDraw();
    output.imageMode(CENTER);
    output.background(0);
    output.endDraw();

  } else if( scale * newscale>=1 && scale * newscale <= maxZoom ) {
    
    scale*=newscale;
    W = WindowW / scale;
    H = WindowH / scale;
    
    outputTemp = output;
    output = createGraphics(W, H);
    
    output.beginDraw();
    output.imageMode(CENTER);
    output.image(outputTemp, W/2, H/2, W, H);
    output.endDraw();
    
  }
  
  outputClipboard = createGraphics(W, H); 
  outputTemp = createGraphics(W, H); 
  
  buffer = new int[W][H];
  screen = createImage(W, H, RGB);
    
  margin = (height-WindowH)/2;
  
  loadPerlinMap();
  
}

 public void screenSave(PGraphics temp) {
  temp.beginDraw();
  temp.imageMode(CENTER);
  temp.image(output, W/2, H/2, W, H);
  temp.endDraw();
  //println("Screen copied.");
}

 public void screenPaste(PGraphics temp) {
  output.beginDraw();
  output.imageMode(CENTER);
  output.image(temp, W/2, H/2, W, H);
  output.endDraw();
  //println("Screen pasted.");
}

 public void clearScreen() {
  output.beginDraw();
  output.background(0);
  output.endDraw();
}

String tempSet;
 public void pauseScreen() {
  if(ruleSet == "(none)") { ruleSet = tempSet; } else { tempSet = ruleSet; ruleSet = "(none)"; }
}

 public void screenSaveToFile() {
  screen.save("screenshots/ca-screenshot-"+year()+"-"+month()+"-"+day()+"-"+hour()+"-"+minute()+"-"+second()+".tif");
  println("Screen saved!");
}
boolean isHovered, isClicked;
int mouseHoldTime;

int bgColor = 235;
int textColor = 100;
int linkColor = 0;
int activeColor = color(0,0,255);
int hoverColor = color(208,225,255);
int pressColor = color(170,205,255);

 public void initConsole() {
  textFont(createFont("Courier",12));
  textSize(12);
  textLeading(12);
}

 public void updateConsole() {
  
  isHovered = false;
  
  int padding = 80;
  
  
  textAlign(RIGHT,BOTTOM); 
  fill(textColor);
  text("FPS: " + round(frameRate), width-20, height-20);
  
  
  textAlign(LEFT,TOP); 
  translate(WindowW+margin+padding, margin);
  
  float titleBarHeight = 50;
  float optionHeight = 25;
  
  pushMatrix();
    fill( modeSet == "RULES" ? activeColor : linkColor);
    text("1  RULES",0,0);
    translate(0, titleBarHeight);
    for(int i=0; i<RULES.length; i++) {
      createHover(-10,-7,150,optionHeight);
      if(isClicked) setRule(RULES[i].name);
      fill( ruleSet == RULES[i].name ? activeColor : linkColor);
      text(RULES[i].shortcut + "  " + RULES[i].name,0,0);
      translate(0,optionHeight);
    }
  popMatrix();
  
  
  translate(90+padding,0);
    
  pushMatrix();
    fill( modeSet == "PALETTES" ? activeColor : linkColor);
    text("2  PALETTES",0,0);
    translate(0, titleBarHeight);
    int palettesCols = 2;
    for(int i=0; i<PALETTES.length; i++) {
      int ix = i % palettesCols * 170;
      int iy = i / palettesCols * 60;
      createHover(ix-10,iy-10,170,60);
      if(isClicked) { setPalette(PALETTES[i].name); PALETTES[i].createThumb(); }
      fill( paletteSet == PALETTES[i].name ? activeColor : linkColor);
      image(PALETTES[i].thumb,ix,iy);
      text(PALETTES[i].name + "\n" + PALETTES[i].shortcut,ix+50,iy+7);
    }
  popMatrix();
  
  
  translate(300+padding,0);
  
  pushMatrix();
    fill( modeSet == "IMAGES" ? activeColor : linkColor);
    text("3  IMAGES",0,0);
    translate(0, titleBarHeight);
    int imagesCols = 4;
    for(int i=0; i<IMAGES.length; i++) {
      int ix = i % imagesCols * 55;
      int iy = i / imagesCols * 68;
      createHover(ix-8,iy-8,55,68);
      if(isClicked) addImage(IMAGES[i].image);
      fill(linkColor);
      image(IMAGES[i].thumb,ix,iy);
      text(IMAGES[i].shortcut,ix,iy+45);
    }
  popMatrix();
  
  
  translate(190+padding,0);
  
  pushMatrix();
    fill( modeSet == "BRUSHES" ? activeColor : linkColor);
    text("4  BRUSHES",0,0);
    translate(0, titleBarHeight);
    for(int i=0; i<BRUSHES.length; i++) {
      createHover(-10,-7,120,optionHeight);
      if(isClicked) setBrush(BRUSHES[i].name);
      fill( brushSet == BRUSHES[i].name ? activeColor : linkColor);
      text(BRUSHES[i].shortcut + "  " + BRUSHES[i].name,0,0);
      translate(0,optionHeight);
    }
    
    translate(0,50);
    fill( modeSet == "GENERATE" ? activeColor : linkColor);
    text("5  GENERATE",0,0);
    translate(0, titleBarHeight);
    for(int i=0; i<GENERATE.length; i++) {
      createHover(-10,-7,120,optionHeight);
      if(isClicked) addElement(GENERATE[i].name);
      fill(linkColor);
      text(GENERATE[i].shortcut + "  " + GENERATE[i].name,0,0);
      translate(0,optionHeight);
    }
  popMatrix();
  
  
  translate(70+padding,0);
  
  
  pushMatrix();
    text("SETTINGS",0,0);
    translate(0, titleBarHeight);
    
    fill(linkColor);
    text("SIZE: " + WindowW/scale+"x"+WindowH/scale, 0, 0);
    translate(0, optionHeight);
    
    for(int i=0; i<7; i++) {
      createHover(30*i-4, -4, 30, 20);
      if(isClicked) { scale = PApplet.parseInt(pow(2,i)); initScreen(1); }
      fill( scale == PApplet.parseInt(pow(2,i)) ? activeColor : linkColor);
      text(PApplet.parseInt(pow(2,9-i)), 30*i, 0);
    }
      
    translate(0, 2*optionHeight);
    
    
    
    createHover(-10,-7,150,optionHeight);
    if(isClicked) { pauseScreen(); }
    fill(linkColor);
    if(ruleSet == "(none)") {
      text("_ START", 0, 0);
    } else {
      text("_ PAUSE", 0, 0);
    }
      
    translate(0, optionHeight);
    
    createHover(-10,-7,150,optionHeight);
    if(isClicked) { clearScreen(); }
    fill(linkColor);
    text("⌫ CLEAR SCREEN", 0, 0);
    
    translate(0, optionHeight);
    
    
    if(automator) {
    
      if(automatorOn) {
        fill(0,255,0);
        text("7 AUTOMATOR ON", 0, 0);
      } else {
        fill(textColor);
        text("7 AUTOMATOR OFF", 0, 0);
      }
    
      fill(textColor);
      translate(0, optionHeight);
      text("8 AUTOMATOR CLEAR", 0, 0);
      translate(0, optionHeight);
      text("9 AUTOMATOR ERASE ", 0, 0);
      
      translate(0, optionHeight);
      
    }
    
    
    createHover(-10,-7,150,optionHeight);
    if(isClicked) { recOn(); }
    if(isRecording) {
      fill(255,0,0);
      text("0 RECORDED: "+framesRecorded, 0, 0);
    } else {
      fill(linkColor);
      text("0 RECORD", 0, 0);
    }
    
    translate(0, optionHeight);
    
    createHover(-10,-7,150,optionHeight);
    if(isClicked) { screenSaveToFile(); }
    fill(linkColor);
    text("⏎ SAVE TO FILE", 0, 0); 
    
    translate(0, optionHeight);
    
    createHover(-10,-7,150,optionHeight);
    if(isClicked) { screenSave(outputClipboard); }
    fill(linkColor);
    text("[ CLIPBOARD COPY", 0, 0); 
    
    translate(0, optionHeight);
    
    createHover(-10,-7,150,optionHeight);
    if(isClicked) { screenPaste(outputClipboard); }
    fill(linkColor);
    text("] CLIPBOARD PASTE", 0, 0); 
    
    translate(0, optionHeight);
    
    createHover(-10,-7,150,optionHeight);
    if(isClicked) { loadPerlinMap(); }
    fill(linkColor);
    text(". RELOARD PERLIN", 0, 0);
    
  popMatrix();
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  if(mousePressed) {
    mouseHoldTime++;
  } else {
    mouseHoldTime=0;
  }
  
  
  cursor( isHovered ? HAND : ARROW );
  
}


 public void createHover(float x, float y, float w, float h) {
  
  boolean hovered = false;
  
  if( mouseX > modelX(0,0,0)+x && mouseX < modelX(0,0,0)+w+x && mouseY > modelY(0,0,0)+y && mouseY < modelY(0,0,0)+h+y ) {
    isHovered = true;
    hovered = true;
  }
  
  isClicked = hovered && ( mouseHoldTime == 1 || mouseHoldTime > 10);
  
  noStroke();
  if(hovered) { fill(hoverColor); }
  if(isClicked) { fill(pressColor); }
  if(hovered) rect(x,y,w,h);
  
}
boolean isRecording = false;
int framesRecorded;
String recUrl;

 public void recOn() {
 isRecording = !isRecording;
 if(isRecording) {
   recUrl = "movies/"+year()+"-"+month()+"-"+day()+"-"+hour()+"-"+minute()+"-"+second()+"/";
   framesRecorded = 0;
 } else {
   println("Movie saved ("+framesRecorded+" frames)");
 }
}

 public void rec() {
 if(isRecording) { 
   screen.save(recUrl + nf(framesRecorded,4) + ".tif");
   framesRecorded++;
 }
}



OscP5 oscP5;
NetAddress LOC;

boolean automatorOn = false;

String excludeChars = "7890-="; 

 public void initAutomator() {
  oscP5 = new OscP5(this,11000);
  LOC = new NetAddress("127.0.0.1",12000);
  clearAutomator();
}

 public void automatorOn() {
  automatorOn = !automatorOn;
}

 public void automatorListen() {
  if( automator && automatorOn ) {
    if( keyPressed && excludeChars.indexOf(key)<0 ) {
      OscMessage MSG = new OscMessage("K");
      MSG.add(key);
      oscP5.send(MSG, LOC);
    } else if(mousePressed) {
      OscMessage MSG = new OscMessage("M");
      MSG.add(mouseX);
      MSG.add(mouseY);
      oscP5.send(MSG, LOC);
    }
  }
}

 public void clearAutomator() {
  OscMessage MSG = new OscMessage("C");
  oscP5.send(MSG, LOC);
}

 public void eraserAutomator() {
  OscMessage MSG = new OscMessage("E");
  oscP5.send(MSG, LOC);
}

 public void oscEvent(OscMessage theOscMessage) {
  char type = theOscMessage.addrPattern().charAt(0);
  if(type=='k') {
    char keyT = theOscMessage.get(0).charValue();
    key = keyT;
    keyPressed();
  }
  if(type=='m') {
    int mouseXT = theOscMessage.get(0).intValue();
    int mouseYT = theOscMessage.get(1).intValue();
    mouseX = mouseXT;
    mouseY = mouseYT;
    mouseHoldTime = 1;
  }
}
Brush[] BRUSHES = {
  new Brush("Rect", 'q'),
  new Brush("Line", 'w'),
  new Brush("Smooth", 'e'),
  new Brush("Spray", 'r')
};

float rect = 80;


/* ----- */


class Brush {
  String name;
  char shortcut;
  Brush(String iname, char ishortcut) {
    name = iname;
    shortcut = ishortcut;
  }
}


/* ----- */


String brushSet;
PImage smoothBrush;
PImage sprayBrush;
int px, py;

 public void loadBrushes() {
  smoothBrush = loadImage("brushes/smooth.png");
  sprayBrush = loadImage("brushes/spray.png");
}

 public void setBrush(String name) {
  brushSet = name;  
}

 public void drawing() {
  
  int x = (mouseX-margin)/scale;
  int y = (mouseY-margin)/scale;
    
  if( (mousePressed||mouseHoldTime>0) && mouseX > margin && mouseX < margin+W*scale && mouseY > margin && mouseY < margin+H*scale) {
  
    
    int c = 255;
    if (mouseButton == RIGHT) { c = 10; }
    output.fill(c);
    
    output.noStroke();
   
    output.beginDraw();
    if(brushSet == "Rect") { output.rect(x-rect*0.375f/scale,y-rect/2/scale,rect*0.75f/scale,rect/scale); }
    if(brushSet == "Line") { output.stroke(c); output.strokeWeight(5/scale); output.line(px,py,x,y); }
    if(brushSet == "Smooth") { output.image(smoothBrush, x, y, 50/scale, 50/scale); }
    if(brushSet == "Spray") { output.image(sprayBrush, x, y, 50/scale, 50/scale); }
    output.endDraw(); 
    
  }
  
   px=x; py=y;
   
}
Generate[] GENERATE = {
  new Generate("Contra", 'q'),
  new Generate("Noise", 'w'),
  new Generate("Snow", 'e'),
  new Generate("Glitch", 'r'),
  new Generate("Flatten", 't')
};


/* ----- */


class Generate {
  String name;
  char shortcut;
  Generate(String iname, char ishortcut) {
    name = iname;
    shortcut = ishortcut;
  }
}


/* ----- */


 public void addElement(String element) {
  
  if(element == "Contra") {
    output.beginDraw();
    int method = PApplet.parseInt(random(0,5));
    if(method==0) {
      int stop = PApplet.parseInt(random(0,W));
      for(int x=0; x < stop; x++) {
        for(int y=0; y < H; y++) {
          output.set(x, y, color( 255-brightness(output.get(x,y)) ) );
        }
      }
    } else if(method==1) {
      int stop = PApplet.parseInt(random(0,W));
      for(int x=stop; x < W; x++) {
        for(int y=0; y < H; y++) {
          output.set(x, y, color( 255-brightness(output.get(x,y)) ) );
        }
      }
    } else if(method==2) {
      int stop = PApplet.parseInt(random(0,H));
      for(int x=0; x < W; x++) {
        for(int y=0; y < stop; y++) {
          output.set(x, y, color( 255-brightness(output.get(x,y)) ) );
        }
      }
    } else if(method==3) {
      int stop = PApplet.parseInt(random(0,H));
      for(int x=0; x < W; x++) {
        for(int y=stop; y < H; y++) {
          output.set(x, y, color( 255-brightness(output.get(x,y)) ) );
        }
      }
    }
    output.endDraw();
  }
  
  
  
  if(element == "Noise") {
    output.beginDraw();
    for(int x=0; x<W; x++) {
      for(int y=0; y<H; y++) {
        //output.set(x, y, int(random(2))*255 );
        output.set(x, y, PApplet.parseInt(random(0,256)) );        
      }
    }
    output.endDraw();
  }
  
  
  
  if(element == "Snow") {
    output.beginDraw();
    for(int x=0; x<500; x++) {
        output.set( PApplet.parseInt(random(0,W)), PApplet.parseInt(random(0,H)), 255 );
    }
    output.endDraw();
  }
  
  
  
  if(element == "Glitch") {
    output.beginDraw();
    
    int count = PApplet.parseInt(random(1,10));
    for(int i=0; i < count; i++) {
      
      int rw = PApplet.parseInt(random(1,300/scale));
      int rh = PApplet.parseInt(random(1,300/scale));
      int sx = PApplet.parseInt(random(0,W-rw));
      int sy = PApplet.parseInt(random(0,H-rh));
      int cx = PApplet.parseInt(random(0,W-rw));
      int cy = PApplet.parseInt(random(0,H-rh));
    
      for(int x=0; x < rw; x++) {
        for(int y=0; y < rh; y++) {
          output.set(sx+x, sy+y, color( brightness(output.get(cx+x,cy+y)) ) );
        }
      }
      
    }
  
  }
  
  if(element == "Flatten") {
    output.beginDraw();
    
    int s = PApplet.parseInt(random(0,256));
    int range = PApplet.parseInt(random(10,50));
    
    for(int x=0; x < W; x++) {
      for(int y=0; y < H; y++) {
        int state = PApplet.parseInt ( brightness(output.get(x,y)) );
        if( state > s-range && state < s+range) {
          output.set(x, y, color(s) );
        }
      }
    }
    
    output.endDraw();
  }
  
  
}
Image[] IMAGES = {
  new Image("images/1.png",'q'),
  new Image("images/2.png",'w'),
  new Image("images/3.png",'e'),
  new Image("images/4.png",'r'),
  new Image("images/5.png",'t'),
  new Image("images/6.png",'a'),
  new Image("images/7.png",'s'),
  new Image("images/8.png",'d'),
  new Image("images/9.png",'f'),
  new Image("images/10.png",'g'),
  new Image("images/11.png",'z'),
  new Image("images/12.png",'x'),
  new Image("images/13.png",'c')
};


/* ----- */


class Image {
  String filename;
  char shortcut;
  PImage thumb;
  PImage image;
  PImage mask;
  Image(String ifilename, char ishortcut) {
    filename = ifilename;
    shortcut = ishortcut;
  }
   public void load() {
    int thumbSize = 40;
    thumb = loadImage(filename);
    thumb.resize(thumbSize, thumbSize);
    image = loadImage(filename);
    image.filter(GRAY);
    image.resize(W, H);
    mask = loadImage(filename);
    mask.filter(THRESHOLD, 0.1f);
    mask.resize(W, H);
    image.mask(mask);
  }
}

 public void loadImages() {
  for(int i=0; i<IMAGES.length; i++) {
    IMAGES[i].load();
  }
}

/* ----- */



 public void addImage(PImage img) {  
  output.beginDraw();
  output.image(img, W/2, H/2, W, H);
  output.endDraw();
}
Palette[] PALETTES = {
  new Palette("Greyscale", 'q'),
  new Palette("Greyscale(R)", 'w'),
  new Palette("Blue", 'e'),
  new Palette("Blue(R)", 'r'),
  new Palette("Minimal(R)", 'a'),
  new Palette("Threshold(R)", 's'),
  new Palette("Stripes", 'd'),
  new Palette("Gas(R)", 'f'),
  new Palette("Color Gas(R)", 'g'),
  new Palette("German Sky", 'c'),
  new Palette("RYB", 'z'),
  new Palette("RYB(R)", 'x'),
  new Palette("Rainbow(R)", 'v')
};


/* ----- */


class Palette {
  String name;
  char shortcut;
  PGraphics thumb;
  Palette(String iname, char ishortcut) {
    name = iname;
    shortcut = ishortcut;
  }
   public void createThumb() {
    setPalette(name);
    int thumbSize = 40;
    thumb = createGraphics(thumbSize,thumbSize);
    thumb.beginDraw();
    thumb.noStroke();
    for(int i=0;i<thumbSize;i++) {
      thumb.fill(palette[ PApplet.parseInt(PApplet.parseFloat(i)/thumbSize*255) ]);
      thumb.rect(0,thumbSize-i,thumbSize,1);
    }
    thumb.endDraw();
  }
}

 public void loadPaletteThumbs() {
  for(int i=0; i<PALETTES.length; i++) {
    PALETTES[i].createThumb();
  }
}


/* ----- */



String paletteSet;

int[] palette = new int[256];
int[] paletteStopsList = new int[256];
int[] paletteStopsPos = new int[256];

 public void setPalette(String name) {
  paletteSet = name;
  colorize();
}
  
 public void colorize() {

  for (int i=0; i<256; i++) { 
    palette[i] = color(0, 0);
    paletteStopsList[i] = color(0, 0);
    paletteStopsPos[i] = 0;
  }

  if (paletteSet=="Greyscale") C_greyscale();
  if (paletteSet=="Greyscale(R)") C_grayscale_R();
  if (paletteSet=="Blue") C_blue();
  if (paletteSet=="Blue(R)") C_blue_R();
  if (paletteSet=="Threshold(R)") C_threshold();
  if (paletteSet=="Rainbow(R)") C_randomRainbow();
  if (paletteSet=="Stripes") C_stripes1();
  if (paletteSet=="Color Gas(R)") C_gasoline1();
  if (paletteSet=="Gas(R)") C_gasoline2();
  if (paletteSet=="RYB") C_ryb();
  if (paletteSet=="RYB(R)") C_amiga();
  if (paletteSet=="Minimal(R)") C_minimal();  
  if (paletteSet=="German Sky") C_germansky();

  addGradients();
}



 public void C_greyscale() {
  for (int i=0; i<256; i++) {
    paletteStopsList[i] = color(i);
  }
}

 public void C_grayscale_R() {
  for (int i=0; i<PApplet.parseInt(random(0,20)); i++) {
    paletteStopsList[PApplet.parseInt(random(0,256))] = color(PApplet.parseInt(random(0,256)));
  }
}

 public void C_blue() {
  paletteStopsList[0] = color(255);
  paletteStopsList[255] = color(0,0,255);
}

 public void C_blue_R() {
  for (int i=0; i<PApplet.parseInt(random(0,20)); i++) {
    paletteStopsList[PApplet.parseInt(random(0,256))] = color(0,0,PApplet.parseInt(random(0,256)));
  }
}




 public void C_minimal() {
  noGradients();
  int count = PApplet.parseInt(random(1, 2));
  for (int i=0; i<count; i++) {
    int stripeW = PApplet.parseInt(random(1, 10));
    int start = PApplet.parseInt(random(0, 256-stripeW));
    for (int j=0; j<stripeW; j++) {
      paletteStopsList[start+j] = color(255);
    }
  }
}

 public void C_amiga() {
  noGradients();
  int[] colors = {
    color(255, 0, 0), 
    color(255, 255, 0), 
    color(0, 0, 255)
  };
  for (int i=0; i < PApplet.parseInt(random(1, 8)); i++) {
    int stripeW = PApplet.parseInt(random(40, 128));
    int start = PApplet.parseInt(random(0, (255-stripeW)));
    int c = colors[PApplet.parseInt(random(0, colors.length))];
    for (int j=0; j < stripeW; j++) {
      paletteStopsList[start+j] = c;
    }
  }
  for (int i=0; i<colors.length; i++) {
    int start = PApplet.parseInt(random(0, 250));
    for (int j=0; j<5; j++) {
      paletteStopsList[start+j] = colors[i];
    }
  }
}

 public void C_threshold() {
  int cent = PApplet.parseInt(random(1, 254));
  paletteStopsList[0] = color(0);
  paletteStopsList[cent] = color(0);
  paletteStopsList[cent+1] = color(255);
  paletteStopsList[255] = color(255);
}

 public void C_stripes1() {
  noGradients();
  for (int i=0; i<4; i++) {
    for (int j=0; j<20; j++) {
      paletteStopsList[i*50+j] = color(255);
    }
  }
}

 public void C_gasoline1() {
  
  int[] colors = {
    color(255, 0, 0), 
    color(255, 255, 0), 
    color(0, 0, 255)
  };
  for (int i=0; i<5; i++) {
    paletteStopsList[i*50] = color(0);
    paletteStopsList[i*50+25] = color(255);
  }
  for (int i=0; i<PApplet.parseInt(random(1, 4)); i++) {
    int ind = PApplet.parseInt(random(0, 205));
    int c = colors[PApplet.parseInt(random(0, colors.length))];
    for (int j=0; j<50; j++) {
      paletteStopsList[ind+j] = c;
    }
  }
}

 public void C_gasoline2() {
  for (int i=0; i<128; i++) { 
    paletteStopsList[i*2] = color(PApplet.parseInt(random(0, 3))*30);
  }
  for (int i=0; i<PApplet.parseInt(random(1, 3)); i++) {
    int ind = PApplet.parseInt(random(0, 225));
    for (int j=0; j<30; j++) {
      paletteStopsList[ind+j] = color(PApplet.parseInt(random(100, 256)));
    }
  }
}

 public void C_randomRainbow() {
  for (int i=0; i<PApplet.parseInt(random(1, 50)); i++) {
    paletteStopsList[PApplet.parseInt(random(0, 256))] = color(PApplet.parseInt(random(0, 256)), PApplet.parseInt(random(0, 256)), PApplet.parseInt(random(0, 256)));
  }
}

 public void C_germansky() {
 paletteStopsList[0] = color(0);
 paletteStopsList[100] = color(255,0,0);
 paletteStopsList[200] = color(255,255,0);
 paletteStopsList[220] = color(118,224,255);
}

 public void C_ryb() {
 paletteStopsList[0] = color(150,0,80);
 paletteStopsList[63] = color(255,0,0);
 
 paletteStopsList[64] = color(255,80,0);
 paletteStopsList[127] = color(255,255,0);
 
 paletteStopsList[128] = color(60,0,100);
 paletteStopsList[191] = color(0,0,255);
 
 paletteStopsList[192] = color(255);
 paletteStopsList[255] = color(255);
}



// GRADIENT OR NOT

 public void noGradients() {
  for (int i=0; i<256; i++) { 
    paletteStopsList[i] = color(0);
  }
}

 public void addGradients() {
  int paletteStopsCount = 0;
  if (alpha(paletteStopsList[0])==0) {
    paletteStopsPos[0]=0; 
    paletteStopsCount++;
  }
  for (int i=0; i<256; i++) { 
    if (alpha(paletteStopsList[i])!=0) {
      paletteStopsPos[paletteStopsCount]=i; 
      paletteStopsCount++;
    }
  }
  if (alpha(paletteStopsList[0])==0) {
    paletteStopsList[0] = paletteStopsList[paletteStopsPos[1]];
  }
  if (alpha(paletteStopsList[255])==0) {
    paletteStopsList[255] = paletteStopsList[paletteStopsPos[paletteStopsCount-1]]; 
    paletteStopsPos[paletteStopsCount]=255; 
    paletteStopsCount++;
  }

  int prevStop=0; 
  int nextStop; 
  int range; 
  int count=0; 

  for (int i=0; i<paletteStopsCount-1; i++) {
    prevStop = paletteStopsPos[i]; 
    nextStop = paletteStopsPos[i+1]; 
    range = nextStop - prevStop; 
    for (int j=0; j<range; j++) {
      float perc = PApplet.parseFloat(j) / range; 
      int r = PApplet.parseInt(    red(paletteStopsList[prevStop]) + perc * ( red(paletteStopsList[nextStop]) - red(paletteStopsList[prevStop]) )    ); 
      int g = PApplet.parseInt(    green(paletteStopsList[prevStop]) + perc * ( green(paletteStopsList[nextStop]) - green(paletteStopsList[prevStop]) )    ); 
      int b = PApplet.parseInt(    blue(paletteStopsList[prevStop]) + perc * ( blue(paletteStopsList[nextStop]) - blue(paletteStopsList[prevStop]) )    ); 
      palette[count] = color(r, g, b); 
      count++;
    }
  }
  palette[255] = paletteStopsList[255];
}
Rule[] RULES = {
  
  new Rule("Blink 1", 'q'),
  new Rule("Blink 2", 'w'),
  new Rule("Blink InOut", 'e'),
  
  new Rule("Perlin", 't'),
  new Rule("Plasma", 'y'),
  new Rule("Plasma Left", 'u'),
  
  new Rule("Expand", 'a'),
  new Rule("Expand RPS", 's'),
    
  new Rule("Dataflow", 'd'),
  new Rule("Feedback", 'f'),
  
  new Rule("Blur", 'z'),
  new Rule("Noise blur", 'x'),
  new Rule("VHS Fadeout", 'c'),  
  new Rule("White shadow", 'v'),  
  new Rule("Sink in", 'b'), 
  
  new Rule("Zipper", 'n'),  
  
  new Rule("Game of Life", 'm')  
  
  //new Rule("New", 'm')
};


/* ----- */


class Rule {
  String name;
  char shortcut;
  Rule(String iname, char ishortcut) {
    name = iname;
    shortcut = ishortcut;
  }
}


/* ----- */


String ruleSet;


float perlinScl;
int[][] perlinMapX;
int[][] perlinMapY;




 public void setRule(String newRule) {
  //if(ruleSet == newRule) { ruleSet = "(none)"; } else { ruleSet = newRule; }
  ruleSet = newRule;
}


 public int checkState(int x, int y, int tx, int ty) {
    
  int state, tempPosX, tempPosY;
  
  if( x+tx < 0 ) { tempPosX = W-1+tx; } else { 
    if( x+tx > W-1 ) { tempPosX = x+tx-W; } else { tempPosX = x+tx; }
  }
  
  if( y+ty < 0 ) { tempPosY = H-1+ty; } else {
    if( y+ty > H-1 ) { tempPosY = y+ty-H; } else { tempPosY = y+ty; }
  }
  
  state = PApplet.parseInt(brightness( output.get( tempPosX, tempPosY ) ));
  
  return state;
}



 public int checkRuleState(int x, int y) {
  
  int nn = checkState(x, y, 0, -1);
  int ss = checkState(x, y, 0, 1);
  int ww = checkState(x, y, -1, 0);
  int ee = checkState(x, y, 1, 0);

  int cc = PApplet.parseInt( brightness(output.get(x, y)) );
  
  int nw = checkState(x, y, -1, -1);
  int ne = checkState(x, y, 1, -1);
  int se = checkState(x, y, 1, 1);
  int sw = checkState(x, y, -1, 1);
  
  int newstate = cc;




  if (ruleSet == "Game of Life") {
    
    int neighbours = 0;
    neighbours += ( nw >=128 ? 1 : 0);
    neighbours += ( nn >=128 ? 1 : 0);
    neighbours += ( ne >=128 ? 1 : 0);
    neighbours += ( ee >=128 ? 1 : 0);
    neighbours += ( se >=128 ? 1 : 0);
    neighbours += ( ss >=128 ? 1 : 0);
    neighbours += ( sw >=128 ? 1 : 0);
    neighbours += ( ww >=128 ? 1 : 0);
    
    if (cc>=128) {
      if (neighbours < 2 || neighbours > 3) {
        newstate = 0;
      }
    } else {  
      if (neighbours == 3 ) {
        newstate = 255;
      }
    }
    
  }
  
  



  if (ruleSet == "Blink 1") {
    float speed = 8;
    newstate = ( newstate < 0 ? 240 : newstate-PApplet.parseInt(speed) );
  }
  
  if (ruleSet == "Blink 2") {
    int neighbours = nw + nn + ne + ee + se + ss + sw + ww;
    if ( newstate > 128 ) {
      if ( neighbours < 72 || neighbours > 108 ) { 
        newstate-=6;
      } else { 
        newstate+=6;
      }
    } else {
      if ( neighbours > 72 && neighbours < 108 ) { 
        newstate+=6;
      } else { 
        newstate-=6;
      }
    }
  }
  
  if (ruleSet == "Blink InOut") {
    if(newstate % 2 == 1) {
      newstate+=2;
    } else {
      newstate-=2;
    }
    if(newstate==255) { newstate=254; } else if (newstate==254) { newstate=255; }
    if(newstate==0) { newstate=1; } else if (newstate==1) { newstate=0; }
  }





  if (ruleSet == "Perlin") {
    newstate = checkState(x, y, perlinMapX[x][y], perlinMapY[x][y] );
  }
  
  if (ruleSet == "Plasma") {
    int[] neighbours = { ss, sw, ww, nw, nn, ne, ee, se };
    int randomN = neighbours[ PApplet.parseInt( constrain(map(noise(x*0.005f,y*0.005f,frameCount*0.05f),0.3f,0.75f,0,1),0,1) * 7 ) ];
    newstate = randomN;
  }
  
  if (ruleSet == "Plasma Left") {
    newstate = checkState(x, y, PApplet.parseInt( noise(x*0.01f,y*0.01f,frameCount*0.02f) * 3 ), 0);
  }
  
  
  
  
  

  if (ruleSet == "Expand") {
    int[] neighbours = { nw, nn, ne, ee, se, ss, sw, ww };
    int randomN = neighbours[PApplet.parseInt(random(0,8))];
    randomN -= 5;
    if(cc == 0) newstate = constrain( randomN, 0, 255 );
  }
  
  if (ruleSet == "Expand RPS") {
    
    int levels = 2;
    // 0 = RED; 1 = GREEN; 2 = BLUE
    int[] neighbours = { nw, nn, ne, ee, se, ss, sw, ww };
    int randomN = neighbours[PApplet.parseInt(random(0,8))];
    int randomC = checkRGB(randomN);
    int currentC = checkRGB(cc);
    
    if( cc > 191 && randomC < 3 && randomN % 64 > 10) { // WHITE
    
      newstate = randomN - levels;
      
    } else {
      
      if( (currentC == 0 && randomC == 2) || (currentC == 1 && randomC == 0) || (currentC == 2 && randomC == 1) ) {
        newstate = newstate + levels;
        newstate = constrain(newstate,0,(currentC+1)*64-1);
      }
      
      if( (currentC == 0 && randomC == 1) || (currentC == 1 && randomC == 2) || (currentC == 2 && randomC == 0) ) {
        newstate = newstate - levels;
        if( newstate < currentC*64) {
          newstate = (randomC+1) * 64 - 1;
        }
      }
      
    }

  }
  
  


  if (ruleSet == "Dataflow") {
    if( cc < 64 ) {
      newstate = nn;
    } else if ( cc < 128 ) {
      newstate = ss;
    } else if ( cc < 192 ) {
      newstate = ee;
    } else {
      newstate = ww;
    }
  }

  if (ruleSet == "Feedback") {
    float scale = 0.95f;
    float margin = (1 - scale) / 2;
    if( x > margin*W && x < (1-margin)*W && y > margin*H && y < (1-margin)*H ) {
      if(newstate<250) {
        float sclr = 1/scale - 1;
        int tx = PApplet.parseInt( PApplet.parseFloat(x-W/2) * sclr );
        int ty = PApplet.parseInt( PApplet.parseFloat(y-H/2) * sclr );
        newstate = 255-checkState(x, y, tx, ty);
      }
    }
  }






  if (ruleSet == "Blur") {
    newstate = ( nw + nn + ne + ee + se + ss + sw + ww + cc ) / 9;
  }

  if (ruleSet == "Noise blur") {
    int range = 10;
    newstate = checkState(x, y, PApplet.parseInt(random(-range,range)), PApplet.parseInt(random(-range,range)) );
  }

  if (ruleSet == "VHS Fadeout") {    
    newstate = (ee + ww) / 2;
  }

  if (ruleSet == "White shadow") {
    if(newstate<230) {      
      if(random(0,100)<1) { newstate += checkState(x, y, PApplet.parseInt(random(-20,5)), PApplet.parseInt(random(-1,1))) * 0.5f ; }
    }
    newstate = constrain(newstate,0,255);
  }
  
  if (ruleSet == "Sink in") {
    if(nn<200) {
      newstate=PApplet.parseInt(nn*0.93f);
    }
  }



  if (ruleSet == "Zipper") {
    newstate = checkState(x, y, (W/2+x)/100, (H/2+y)/100 );
  }
  
  
  
  
  
  
  if (ruleSet == "New") { 
  }
  
  return newstate;
  
}

 public int checkRGB(int state) {
  if(state<=63) { return 0; } else if(state<=127) { return 1; } else if(state<=191) { return 2; } else { return 3; }
}


 public void loadPerlinMap() {
  float scl = random(0.0005f,0.006f) * scale;
  int range = PApplet.parseInt(random(4,20)) / scale;
  if(range<5) range=5;
  perlinMapX = new int[W][H];
  perlinMapY = new int[W][H];
  float startPointX = random(0,1);
  float startPointY = random(0,1);
  for(int a=0; a<W; a++) {
    for(int b=0; b<H; b++) {
      perlinMapX[a][b] = PApplet.parseInt(map(noise(0,startPointX+b*scl),0,1,-range,range));
      perlinMapY[a][b] = PApplet.parseInt(map(noise(startPointY/2+a*scl,0),0,1,-range,range));
    }
  }
}


  public void settings() { size(1920, 700, P3D);
noSmooth(); }

  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "CELLULAR_AUTOMATA" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
