Palette[] PALETTES = {
  new Palette("Greyscale", 'q'),
  new Palette("Greyscale(R)", 'w'),
  new Palette("Blue", 'e'),
  new Palette("Blue(R)", 'r'),
  new Palette("Minimal(R)", 'a'),
  new Palette("Threshold(R)", 's'),
  new Palette("Stripes", 'd'),
  new Palette("Gas(R)", 'f'),
  new Palette("Color Gas(R)", 'g'),
  new Palette("German Sky", 'c'),
  new Palette("RYB", 'z'),
  new Palette("RYB(R)", 'x'),
  new Palette("Rainbow(R)", 'v')
};


/* ----- */


class Palette {
  String name;
  char shortcut;
  PGraphics thumb;
  Palette(String iname, char ishortcut) {
    name = iname;
    shortcut = ishortcut;
  }
  void createThumb() {
    setPalette(name);
    int thumbSize = 40;
    thumb = createGraphics(thumbSize,thumbSize);
    thumb.beginDraw();
    thumb.noStroke();
    for(int i=0;i<thumbSize;i++) {
      thumb.fill(palette[ int(float(i)/thumbSize*255) ]);
      thumb.rect(0,thumbSize-i,thumbSize,1);
    }
    thumb.endDraw();
  }
}

void loadPaletteThumbs() {
  for(int i=0; i<PALETTES.length; i++) {
    PALETTES[i].createThumb();
  }
}


/* ----- */



String paletteSet;

color[] palette = new color[256];
color[] paletteStopsList = new color[256];
int[] paletteStopsPos = new int[256];

void setPalette(String name) {
  paletteSet = name;
  colorize();
}
  
void colorize() {

  for (int i=0; i<256; i++) { 
    palette[i] = color(0, 0);
    paletteStopsList[i] = color(0, 0);
    paletteStopsPos[i] = 0;
  }

  if (paletteSet=="Greyscale") C_greyscale();
  if (paletteSet=="Greyscale(R)") C_grayscale_R();
  if (paletteSet=="Blue") C_blue();
  if (paletteSet=="Blue(R)") C_blue_R();
  if (paletteSet=="Threshold(R)") C_threshold();
  if (paletteSet=="Rainbow(R)") C_randomRainbow();
  if (paletteSet=="Stripes") C_stripes1();
  if (paletteSet=="Color Gas(R)") C_gasoline1();
  if (paletteSet=="Gas(R)") C_gasoline2();
  if (paletteSet=="RYB") C_ryb();
  if (paletteSet=="RYB(R)") C_amiga();
  if (paletteSet=="Minimal(R)") C_minimal();  
  if (paletteSet=="German Sky") C_germansky();

  addGradients();
}



void C_greyscale() {
  for (int i=0; i<256; i++) {
    paletteStopsList[i] = color(i);
  }
}

void C_grayscale_R() {
  for (int i=0; i<int(random(0,20)); i++) {
    paletteStopsList[int(random(0,256))] = color(int(random(0,256)));
  }
}

void C_blue() {
  paletteStopsList[0] = color(255);
  paletteStopsList[255] = color(0,0,255);
}

void C_blue_R() {
  for (int i=0; i<int(random(0,20)); i++) {
    paletteStopsList[int(random(0,256))] = color(0,0,int(random(0,256)));
  }
}




void C_minimal() {
  noGradients();
  int count = int(random(1, 2));
  for (int i=0; i<count; i++) {
    int stripeW = int(random(1, 10));
    int start = int(random(0, 256-stripeW));
    for (int j=0; j<stripeW; j++) {
      paletteStopsList[start+j] = color(255);
    }
  }
}

void C_amiga() {
  noGradients();
  color[] colors = {
    color(255, 0, 0), 
    color(255, 255, 0), 
    color(0, 0, 255)
  };
  for (int i=0; i < int(random(1, 8)); i++) {
    int stripeW = int(random(40, 128));
    int start = int(random(0, (255-stripeW)));
    color c = colors[int(random(0, colors.length))];
    for (int j=0; j < stripeW; j++) {
      paletteStopsList[start+j] = c;
    }
  }
  for (int i=0; i<colors.length; i++) {
    int start = int(random(0, 250));
    for (int j=0; j<5; j++) {
      paletteStopsList[start+j] = colors[i];
    }
  }
}

void C_threshold() {
  int cent = int(random(1, 254));
  paletteStopsList[0] = color(0);
  paletteStopsList[cent] = color(0);
  paletteStopsList[cent+1] = color(255);
  paletteStopsList[255] = color(255);
}

void C_stripes1() {
  noGradients();
  for (int i=0; i<4; i++) {
    for (int j=0; j<20; j++) {
      paletteStopsList[i*50+j] = color(255);
    }
  }
}

void C_gasoline1() {
  
  color[] colors = {
    color(255, 0, 0), 
    color(255, 255, 0), 
    color(0, 0, 255)
  };
  for (int i=0; i<5; i++) {
    paletteStopsList[i*50] = color(0);
    paletteStopsList[i*50+25] = color(255);
  }
  for (int i=0; i<int(random(1, 4)); i++) {
    int ind = int(random(0, 205));
    color c = colors[int(random(0, colors.length))];
    for (int j=0; j<50; j++) {
      paletteStopsList[ind+j] = c;
    }
  }
}

void C_gasoline2() {
  for (int i=0; i<128; i++) { 
    paletteStopsList[i*2] = color(int(random(0, 3))*30);
  }
  for (int i=0; i<int(random(1, 3)); i++) {
    int ind = int(random(0, 225));
    for (int j=0; j<30; j++) {
      paletteStopsList[ind+j] = color(int(random(100, 256)));
    }
  }
}

void C_randomRainbow() {
  for (int i=0; i<int(random(1, 50)); i++) {
    paletteStopsList[int(random(0, 256))] = color(int(random(0, 256)), int(random(0, 256)), int(random(0, 256)));
  }
}

void C_germansky() {
 paletteStopsList[0] = color(0);
 paletteStopsList[100] = color(255,0,0);
 paletteStopsList[200] = color(255,255,0);
 paletteStopsList[220] = color(118,224,255);
}

void C_ryb() {
 paletteStopsList[0] = color(150,0,80);
 paletteStopsList[63] = color(255,0,0);
 
 paletteStopsList[64] = color(255,80,0);
 paletteStopsList[127] = color(255,255,0);
 
 paletteStopsList[128] = color(60,0,100);
 paletteStopsList[191] = color(0,0,255);
 
 paletteStopsList[192] = color(255);
 paletteStopsList[255] = color(255);
}



// GRADIENT OR NOT

void noGradients() {
  for (int i=0; i<256; i++) { 
    paletteStopsList[i] = color(0);
  }
}

void addGradients() {
  int paletteStopsCount = 0;
  if (alpha(paletteStopsList[0])==0) {
    paletteStopsPos[0]=0; 
    paletteStopsCount++;
  }
  for (int i=0; i<256; i++) { 
    if (alpha(paletteStopsList[i])!=0) {
      paletteStopsPos[paletteStopsCount]=i; 
      paletteStopsCount++;
    }
  }
  if (alpha(paletteStopsList[0])==0) {
    paletteStopsList[0] = paletteStopsList[paletteStopsPos[1]];
  }
  if (alpha(paletteStopsList[255])==0) {
    paletteStopsList[255] = paletteStopsList[paletteStopsPos[paletteStopsCount-1]]; 
    paletteStopsPos[paletteStopsCount]=255; 
    paletteStopsCount++;
  }

  int prevStop=0; 
  int nextStop; 
  int range; 
  int count=0; 

  for (int i=0; i<paletteStopsCount-1; i++) {
    prevStop = paletteStopsPos[i]; 
    nextStop = paletteStopsPos[i+1]; 
    range = nextStop - prevStop; 
    for (int j=0; j<range; j++) {
      float perc = float(j) / range; 
      int r = int(    red(paletteStopsList[prevStop]) + perc * ( red(paletteStopsList[nextStop]) - red(paletteStopsList[prevStop]) )    ); 
      int g = int(    green(paletteStopsList[prevStop]) + perc * ( green(paletteStopsList[nextStop]) - green(paletteStopsList[prevStop]) )    ); 
      int b = int(    blue(paletteStopsList[prevStop]) + perc * ( blue(paletteStopsList[nextStop]) - blue(paletteStopsList[prevStop]) )    ); 
      palette[count] = color(r, g, b); 
      count++;
    }
  }
  palette[255] = paletteStopsList[255];
}
