Rule[] RULES = {
  
  new Rule("Blink 1", 'q'),
  new Rule("Blink 2", 'w'),
  new Rule("Blink InOut", 'e'),
  
  new Rule("Perlin", 't'),
  new Rule("Plasma", 'y'),
  new Rule("Plasma Left", 'u'),
  
  new Rule("Expand", 'a'),
  new Rule("Expand RPS", 's'),
    
  new Rule("Dataflow", 'd'),
  new Rule("Feedback", 'f'),
  
  new Rule("Blur", 'z'),
  new Rule("Noise blur", 'x'),
  new Rule("VHS Fadeout", 'c'),  
  new Rule("White shadow", 'v'),  
  new Rule("Sink in", 'b'), 
  
  new Rule("Zipper", 'n'),  
  
  new Rule("Game of Life", 'm')  
  
  //new Rule("New", 'm')
};


/* ----- */


class Rule {
  String name;
  char shortcut;
  Rule(String iname, char ishortcut) {
    name = iname;
    shortcut = ishortcut;
  }
}


/* ----- */


String ruleSet;


float perlinScl;
int[][] perlinMapX;
int[][] perlinMapY;




void setRule(String newRule) {
  //if(ruleSet == newRule) { ruleSet = "(none)"; } else { ruleSet = newRule; }
  ruleSet = newRule;
}


int checkState(int x, int y, int tx, int ty) {
    
  int state, tempPosX, tempPosY;
  
  if( x+tx < 0 ) { tempPosX = W-1+tx; } else { 
    if( x+tx > W-1 ) { tempPosX = x+tx-W; } else { tempPosX = x+tx; }
  }
  
  if( y+ty < 0 ) { tempPosY = H-1+ty; } else {
    if( y+ty > H-1 ) { tempPosY = y+ty-H; } else { tempPosY = y+ty; }
  }
  
  state = int(brightness( output.get( tempPosX, tempPosY ) ));
  
  return state;
}



int checkRuleState(int x, int y) {
  
  int nn = checkState(x, y, 0, -1);
  int ss = checkState(x, y, 0, 1);
  int ww = checkState(x, y, -1, 0);
  int ee = checkState(x, y, 1, 0);

  int cc = int( brightness(output.get(x, y)) );
  
  int nw = checkState(x, y, -1, -1);
  int ne = checkState(x, y, 1, -1);
  int se = checkState(x, y, 1, 1);
  int sw = checkState(x, y, -1, 1);
  
  int newstate = cc;




  if (ruleSet == "Game of Life") {
    
    int neighbours = 0;
    neighbours += ( nw >=128 ? 1 : 0);
    neighbours += ( nn >=128 ? 1 : 0);
    neighbours += ( ne >=128 ? 1 : 0);
    neighbours += ( ee >=128 ? 1 : 0);
    neighbours += ( se >=128 ? 1 : 0);
    neighbours += ( ss >=128 ? 1 : 0);
    neighbours += ( sw >=128 ? 1 : 0);
    neighbours += ( ww >=128 ? 1 : 0);
    
    if (cc>=128) {
      if (neighbours < 2 || neighbours > 3) {
        newstate = 0;
      }
    } else {  
      if (neighbours == 3 ) {
        newstate = 255;
      }
    }
    
  }
  
  



  if (ruleSet == "Blink 1") {
    float speed = 8;
    newstate = ( newstate < 0 ? 240 : newstate-int(speed) );
  }
  
  if (ruleSet == "Blink 2") {
    int neighbours = nw + nn + ne + ee + se + ss + sw + ww;
    if ( newstate > 128 ) {
      if ( neighbours < 72 || neighbours > 108 ) { 
        newstate-=6;
      } else { 
        newstate+=6;
      }
    } else {
      if ( neighbours > 72 && neighbours < 108 ) { 
        newstate+=6;
      } else { 
        newstate-=6;
      }
    }
  }
  
  if (ruleSet == "Blink InOut") {
    if(newstate % 2 == 1) {
      newstate+=2;
    } else {
      newstate-=2;
    }
    if(newstate==255) { newstate=254; } else if (newstate==254) { newstate=255; }
    if(newstate==0) { newstate=1; } else if (newstate==1) { newstate=0; }
  }





  if (ruleSet == "Perlin") {
    newstate = checkState(x, y, perlinMapX[x][y], perlinMapY[x][y] );
  }
  
  if (ruleSet == "Plasma") {
    int[] neighbours = { ss, sw, ww, nw, nn, ne, ee, se };
    int randomN = neighbours[ int( constrain(map(noise(x*0.005,y*0.005,frameCount*0.05),0.3,0.75,0,1),0,1) * 7 ) ];
    newstate = randomN;
  }
  
  if (ruleSet == "Plasma Left") {
    newstate = checkState(x, y, int( noise(x*0.01,y*0.01,frameCount*0.02) * 3 ), 0);
  }
  
  
  
  
  

  if (ruleSet == "Expand") {
    int[] neighbours = { nw, nn, ne, ee, se, ss, sw, ww };
    int randomN = neighbours[int(random(0,8))];
    randomN -= 5;
    if(cc == 0) newstate = constrain( randomN, 0, 255 );
  }
  
  if (ruleSet == "Expand RPS") {
    
    int levels = 2;
    // 0 = RED; 1 = GREEN; 2 = BLUE
    int[] neighbours = { nw, nn, ne, ee, se, ss, sw, ww };
    int randomN = neighbours[int(random(0,8))];
    int randomC = checkRGB(randomN);
    int currentC = checkRGB(cc);
    
    if( cc > 191 && randomC < 3 && randomN % 64 > 10) { // WHITE
    
      newstate = randomN - levels;
      
    } else {
      
      if( (currentC == 0 && randomC == 2) || (currentC == 1 && randomC == 0) || (currentC == 2 && randomC == 1) ) {
        newstate = newstate + levels;
        newstate = constrain(newstate,0,(currentC+1)*64-1);
      }
      
      if( (currentC == 0 && randomC == 1) || (currentC == 1 && randomC == 2) || (currentC == 2 && randomC == 0) ) {
        newstate = newstate - levels;
        if( newstate < currentC*64) {
          newstate = (randomC+1) * 64 - 1;
        }
      }
      
    }

  }
  
  


  if (ruleSet == "Dataflow") {
    if( cc < 64 ) {
      newstate = nn;
    } else if ( cc < 128 ) {
      newstate = ss;
    } else if ( cc < 192 ) {
      newstate = ee;
    } else {
      newstate = ww;
    }
  }

  if (ruleSet == "Feedback") {
    float scale = 0.95;
    float margin = (1 - scale) / 2;
    if( x > margin*W && x < (1-margin)*W && y > margin*H && y < (1-margin)*H ) {
      if(newstate<250) {
        float sclr = 1/scale - 1;
        int tx = int( float(x-W/2) * sclr );
        int ty = int( float(y-H/2) * sclr );
        newstate = 255-checkState(x, y, tx, ty);
      }
    }
  }






  if (ruleSet == "Blur") {
    newstate = ( nw + nn + ne + ee + se + ss + sw + ww + cc ) / 9;
  }

  if (ruleSet == "Noise blur") {
    int range = 10;
    newstate = checkState(x, y, int(random(-range,range)), int(random(-range,range)) );
  }

  if (ruleSet == "VHS Fadeout") {    
    newstate = (ee + ww) / 2;
  }

  if (ruleSet == "White shadow") {
    if(newstate<230) {      
      if(random(0,100)<1) { newstate += checkState(x, y, int(random(-20,5)), int(random(-1,1))) * 0.5 ; }
    }
    newstate = constrain(newstate,0,255);
  }
  
  if (ruleSet == "Sink in") {
    if(nn<200) {
      newstate=int(nn*0.93);
    }
  }



  if (ruleSet == "Zipper") {
    newstate = checkState(x, y, (W/2+x)/100, (H/2+y)/100 );
  }
  
  
  
  
  
  
  if (ruleSet == "New") { 
  }
  
  return newstate;
  
}

int checkRGB(int state) {
  if(state<=63) { return 0; } else if(state<=127) { return 1; } else if(state<=191) { return 2; } else { return 3; }
}


void loadPerlinMap() {
  float scl = random(0.0005,0.006) * scale;
  int range = int(random(4,20)) / scale;
  if(range<5) range=5;
  perlinMapX = new int[W][H];
  perlinMapY = new int[W][H];
  float startPointX = random(0,1);
  float startPointY = random(0,1);
  for(int a=0; a<W; a++) {
    for(int b=0; b<H; b++) {
      perlinMapX[a][b] = int(map(noise(0,startPointX+b*scl),0,1,-range,range));
      perlinMapY[a][b] = int(map(noise(startPointY/2+a*scl,0),0,1,-range,range));
    }
  }
}
